@@ -3,23 +3,24 @@
  * Copyright © 2011-2012 Collabora, Ltd.
  * Copyright © 2013 Raspberry Pi Foundation
  *
- * Permission to use, copy, modify, distribute, and sell this software and
- * its documentation for any purpose is hereby granted without fee, provided
- * that the above copyright notice appear in all copies and that both that
- * copyright notice and this permission notice appear in supporting
- * documentation, and that the name of the copyright holders not be used in
- * advertising or publicity pertaining to distribution of the software
- * without specific, written prior permission.  The copyright holders make
- * no representations about the suitability of this software for any
- * purpose.  It is provided "as is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
- * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
- * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
- * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
- * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
- * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
- * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
 #include "config.h"
@@ -29,8 +30,8 @@
 #include <string.h>
 
 #include "shell.h"
-#include "desktop-shell-server-protocol.h"
-#include "input-method-server-protocol.h"
+#include "input-method-unstable-v1-server-protocol.h"
+#include "shared/helpers.h"
 
 struct input_panel_surface {
 	struct wl_resource *resource;
@@ -66,17 +67,22 @@ show_input_panel_surface(struct input_panel_surface *ipsurf)
 	float x, y;
 
 	wl_list_for_each(seat, &shell->compositor->seat_list, link) {
-		if (!seat->keyboard)
+		struct weston_keyboard *keyboard =
+			weston_seat_get_keyboard(seat);
+
+		if (!keyboard || !keyboard->focus)
+			continue;
+		focus = weston_surface_get_main_surface(keyboard->focus);
+		if (!focus)
 			continue;
-		focus = weston_surface_get_main_surface(seat->keyboard->focus);
 		ipsurf->output = focus->output;
 		x = ipsurf->output->x + (ipsurf->output->width - ipsurf->surface->width) / 2;
 		y = ipsurf->output->y + ipsurf->output->height - ipsurf->surface->height;
 		weston_view_set_position(ipsurf->view, x, y);
 	}
 
-	wl_list_insert(&shell->input_panel_layer.view_list,
-		       &ipsurf->view->layer_link);
+	weston_layer_entry_insert(&shell->input_panel_layer.view_list,
+	                          &ipsurf->view->layer_link);
 	weston_view_geometry_dirty(ipsurf->view);
 	weston_view_update_transform(ipsurf->view);
 	weston_surface_damage(ipsurf->surface);
@@ -135,7 +141,8 @@ hide_input_panels(struct wl_listener *listener, void *data)
 		wl_list_remove(&shell->input_panel_layer.link);
 
 	wl_list_for_each_safe(view, next,
-			      &shell->input_panel_layer.view_list, layer_link)
+			      &shell->input_panel_layer.view_list.link,
+			      layer_link.link)
 		weston_view_unmap(view);
 }
 
@@ -149,6 +156,12 @@ update_input_panels(struct wl_listener *listener, void *data)
 	memcpy(&shell->text_input.cursor_rectangle, data, sizeof(pixman_box32_t));
 }
 
+static int
+input_panel_get_label(struct weston_surface *surface, char *buf, size_t len)
+{
+	return snprintf(buf, len, "input panel");
+}
+
 static void
 input_panel_configure(struct weston_surface *surface, int32_t sx, int32_t sy)
 {
@@ -186,6 +199,7 @@ destroy_input_panel_surface(struct input_panel_surface *input_panel_surface)
 	wl_list_remove(&input_panel_surface->link);
 
 	input_panel_surface->surface->configure = NULL;
+	weston_surface_set_label_func(input_panel_surface->surface, NULL);
 	weston_view_destroy(input_panel_surface->view);
 
 	free(input_panel_surface);
@@ -227,6 +241,7 @@ create_input_panel_surface(struct desktop_shell *shell,
 
 	surface->configure = input_panel_configure;
 	surface->configure_private = input_panel_surface;
+	weston_surface_set_label_func(surface, input_panel_get_label);
 
 	input_panel_surface->shell = shell;
 
@@ -274,7 +289,7 @@ input_panel_surface_set_overlay_panel(struct wl_client *client,
 	input_panel_surface->panel = 1;
 }
 
-static const struct wl_input_panel_surface_interface input_panel_surface_implementation = {
+static const struct zwp_input_panel_surface_v1_interface input_panel_surface_implementation = {
 	input_panel_surface_set_toplevel,
 	input_panel_surface_set_overlay_panel
 };
@@ -316,14 +331,16 @@ input_panel_get_input_panel_surface(struct wl_client *client,
 
 	ipsurf->resource =
 		wl_resource_create(client,
-				   &wl_input_panel_surface_interface, 1, id);
+				   &zwp_input_panel_surface_v1_interface,
+				   1,
+				   id);
 	wl_resource_set_implementation(ipsurf->resource,
 				       &input_panel_surface_implementation,
 				       ipsurf,
 				       destroy_input_panel_surface_resource);
 }
 
-static const struct wl_input_panel_interface input_panel_implementation = {
+static const struct zwp_input_panel_v1_interface input_panel_implementation = {
 	input_panel_get_input_panel_surface
 };
 
@@ -343,7 +360,7 @@ bind_input_panel(struct wl_client *client,
 	struct wl_resource *resource;
 
 	resource = wl_resource_create(client,
-				      &wl_input_panel_interface, 1, id);
+				      &zwp_input_panel_v1_interface, 1, id);
 
 	if (shell->input_panel.binding == NULL) {
 		wl_resource_set_implementation(resource,
@@ -355,7 +372,6 @@ bind_input_panel(struct wl_client *client,
 
 	wl_resource_post_error(resource, WL_DISPLAY_ERROR_INVALID_OBJECT,
 			       "interface object already bound");
-	wl_resource_destroy(resource);
 }
 
 void
@@ -383,7 +399,7 @@ input_panel_setup(struct desktop_shell *shell)
 	wl_list_init(&shell->input_panel.surfaces);
 
 	if (wl_global_create(shell->compositor->wl_display,
-			     &wl_input_panel_interface, 1,
+			     &zwp_input_panel_v1_interface, 1,
 			     shell, bind_input_panel) == NULL)
 		return -1;
 
