@@ -1,23 +1,26 @@
 /*
  * Copyright © 2011 Kristian Høgsberg
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include "config.h"
@@ -29,6 +32,7 @@
 #include <assert.h>
 
 #include "compositor.h"
+#include "shared/helpers.h"
 
 struct weston_drag {
 	struct wl_client *client;
@@ -40,6 +44,7 @@ struct weston_drag {
 	struct weston_view *icon;
 	struct wl_listener icon_destroy_listener;
 	int32_t dx, dy;
+	struct weston_keyboard_grab keyboard_grab;
 };
 
 struct weston_pointer_drag {
@@ -52,12 +57,9 @@ struct weston_touch_drag {
 	struct weston_touch_grab grab;
 };
 
-static void
-empty_region(pixman_region32_t *region)
-{
-	pixman_region32_fini(region);
-	pixman_region32_init(region);
-}
+#define ALL_ACTIONS (WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY | \
+		     WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE | \
+		     WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK)
 
 static void
 data_offer_accept(struct wl_client *client, struct wl_resource *resource,
@@ -65,12 +67,16 @@ data_offer_accept(struct wl_client *client, struct wl_resource *resource,
 {
 	struct weston_data_offer *offer = wl_resource_get_user_data(resource);
 
+	/* Protect against untimely calls from older data offers */
+	if (!offer->source || offer != offer->source->offer)
+		return;
+
 	/* FIXME: Check that client is currently focused by the input
 	 * device that is currently dragging this data source.  Should
 	 * this be a wl_data_device request? */
 
-	if (offer->source)
-		offer->source->accept(offer->source, serial, mime_type);
+	offer->source->accept(offer->source, serial, mime_type);
+	offer->source->accepted = mime_type != NULL;
 }
 
 static void
@@ -79,7 +85,7 @@ data_offer_receive(struct wl_client *client, struct wl_resource *resource,
 {
 	struct weston_data_offer *offer = wl_resource_get_user_data(resource);
 
-	if (offer->source)
+	if (offer->source && offer == offer->source->offer)
 		offer->source->send(offer->source, mime_type, fd);
 	else
 		close(fd);
@@ -91,10 +97,158 @@ data_offer_destroy(struct wl_client *client, struct wl_resource *resource)
 	wl_resource_destroy(resource);
 }
 
+static void
+data_source_notify_finish(struct weston_data_source *source)
+{
+	if (!source->actions_set)
+		return;
+
+	if (source->offer->in_ask &&
+	    wl_resource_get_version(source->resource) >=
+	    WL_DATA_SOURCE_ACTION_SINCE_VERSION) {
+		wl_data_source_send_action(source->resource,
+					   source->current_dnd_action);
+	}
+
+	if (wl_resource_get_version(source->resource) >=
+	    WL_DATA_SOURCE_DND_FINISHED_SINCE_VERSION) {
+		wl_data_source_send_dnd_finished(source->resource);
+	}
+
+	source->offer = NULL;
+}
+
+static uint32_t
+data_offer_choose_action(struct weston_data_offer *offer)
+{
+	uint32_t available_actions, preferred_action = 0;
+	uint32_t source_actions, offer_actions;
+
+	if (wl_resource_get_version(offer->resource) >=
+	    WL_DATA_OFFER_ACTION_SINCE_VERSION) {
+		offer_actions = offer->dnd_actions;
+		preferred_action = offer->preferred_dnd_action;
+	} else {
+		offer_actions = WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY;
+	}
+
+	if (wl_resource_get_version(offer->source->resource) >=
+	    WL_DATA_SOURCE_ACTION_SINCE_VERSION)
+		source_actions = offer->source->dnd_actions;
+	else
+		source_actions = WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY;
+
+	available_actions = offer_actions & source_actions;
+
+	if (!available_actions)
+		return WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
+
+	if (offer->source->seat &&
+	    offer->source->compositor_action & available_actions)
+		return offer->source->compositor_action;
+
+	/* If the dest side has a preferred DnD action, use it */
+	if ((preferred_action & available_actions) != 0)
+		return preferred_action;
+
+	/* Use the first found action, in bit order */
+	return 1 << (ffs(available_actions) - 1);
+}
+
+static void
+data_offer_update_action(struct weston_data_offer *offer)
+{
+	uint32_t action;
+
+	if (!offer->source)
+		return;
+
+	action = data_offer_choose_action(offer);
+
+	if (offer->source->current_dnd_action == action)
+		return;
+
+	offer->source->current_dnd_action = action;
+
+	if (offer->in_ask)
+		return;
+
+	if (wl_resource_get_version(offer->source->resource) >=
+	    WL_DATA_SOURCE_ACTION_SINCE_VERSION)
+		wl_data_source_send_action(offer->source->resource, action);
+
+	if (wl_resource_get_version(offer->resource) >=
+	    WL_DATA_OFFER_ACTION_SINCE_VERSION)
+		wl_data_offer_send_action(offer->resource, action);
+}
+
+static void
+data_offer_set_actions(struct wl_client *client,
+		       struct wl_resource *resource,
+		       uint32_t dnd_actions, uint32_t preferred_action)
+{
+	struct weston_data_offer *offer = wl_resource_get_user_data(resource);
+
+	if (dnd_actions & ~ALL_ACTIONS) {
+		wl_resource_post_error(offer->resource,
+				       WL_DATA_OFFER_ERROR_INVALID_ACTION_MASK,
+				       "invalid action mask %x", dnd_actions);
+		return;
+	}
+
+	if (preferred_action &&
+	    (!(preferred_action & dnd_actions) ||
+	     __builtin_popcount(preferred_action) > 1)) {
+		wl_resource_post_error(offer->resource,
+				       WL_DATA_OFFER_ERROR_INVALID_ACTION,
+				       "invalid action %x", preferred_action);
+		return;
+	}
+
+	offer->dnd_actions = dnd_actions;
+	offer->preferred_dnd_action = preferred_action;
+	data_offer_update_action(offer);
+}
+
+static void
+data_offer_finish(struct wl_client *client, struct wl_resource *resource)
+{
+	struct weston_data_offer *offer = wl_resource_get_user_data(resource);
+
+	if (!offer->source || offer->source->offer != offer)
+		return;
+
+	/* Disallow finish while we have a grab driving drag-and-drop, or
+	 * if the negotiation is not at the right stage
+	 */
+	if (offer->source->seat ||
+	    !offer->source->accepted) {
+		wl_resource_post_error(offer->resource,
+				       WL_DATA_OFFER_ERROR_INVALID_FINISH,
+				       "premature finish request");
+		return;
+	}
+
+	switch (offer->source->current_dnd_action) {
+	case WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE:
+	case WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK:
+		wl_resource_post_error(offer->resource,
+				       WL_DATA_OFFER_ERROR_INVALID_OFFER,
+				       "offer finished with an invalid action");
+		return;
+	default:
+		break;
+	}
+
+	data_source_notify_finish(offer->source);
+}
+
 static const struct wl_data_offer_interface data_offer_interface = {
 	data_offer_accept,
 	data_offer_receive,
 	data_offer_destroy,
+	data_offer_finish,
+	data_offer_set_actions,
 };
 
 static void
@@ -102,8 +256,29 @@ destroy_data_offer(struct wl_resource *resource)
 {
 	struct weston_data_offer *offer = wl_resource_get_user_data(resource);
 
-	if (offer->source)
-		wl_list_remove(&offer->source_destroy_listener.link);
+	if (!offer->source)
+		goto out;
+
+	wl_list_remove(&offer->source_destroy_listener.link);
+
+	if (offer->source->offer != offer)
+		goto out;
+
+	/* If the drag destination has version < 3, wl_data_offer.finish
+	 * won't be called, so do this here as a safety net, because
+	 * we still want the version >=3 drag source to be happy.
+	 */
+	if (wl_resource_get_version(offer->resource) <
+	    WL_DATA_OFFER_ACTION_SINCE_VERSION) {
+		data_source_notify_finish(offer->source);
+	} else if (offer->source->resource &&
+		   wl_resource_get_version(offer->source->resource) >=
+		   WL_DATA_SOURCE_DND_FINISHED_SINCE_VERSION) {
+		wl_data_source_send_cancelled(offer->source->resource);
+	}
+
+	offer->source->offer = NULL;
+out:
 	free(offer);
 }
 
@@ -118,7 +293,7 @@ destroy_offer_data_source(struct wl_listener *listener, void *data)
 	offer->source = NULL;
 }
 
-static struct wl_resource *
+static struct weston_data_offer *
 weston_data_source_send_offer(struct weston_data_source *source,
 			      struct wl_resource *target)
 {
@@ -131,7 +306,8 @@ weston_data_source_send_offer(struct weston_data_source *source,
 
 	offer->resource =
 		wl_resource_create(wl_resource_get_client(target),
-				   &wl_data_offer_interface, 1, 0);
+				   &wl_data_offer_interface,
+				   wl_resource_get_version(target), 0);
 	if (offer->resource == NULL) {
 		free(offer);
 		return NULL;
@@ -140,6 +316,9 @@ weston_data_source_send_offer(struct weston_data_source *source,
 	wl_resource_set_implementation(offer->resource, &data_offer_interface,
 				       offer, destroy_data_offer);
 
+	offer->in_ask = false;
+	offer->dnd_actions = 0;
+	offer->preferred_dnd_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
 	offer->source = source;
 	offer->source_destroy_listener.notify = destroy_offer_data_source;
 	wl_signal_add(&source->destroy_signal,
@@ -150,7 +329,10 @@ weston_data_source_send_offer(struct weston_data_source *source,
 	wl_array_for_each(p, &source->mime_types)
 		wl_data_offer_send_offer(offer->resource, *p);
 
-	return offer->resource;
+	source->offer = offer;
+	source->accepted = false;
+
+	return offer;
 }
 
 static void
@@ -175,9 +357,44 @@ data_source_destroy(struct wl_client *client, struct wl_resource *resource)
 	wl_resource_destroy(resource);
 }
 
+static void
+data_source_set_actions(struct wl_client *client,
+			struct wl_resource *resource,
+			uint32_t dnd_actions)
+{
+	struct weston_data_source *source =
+		wl_resource_get_user_data(resource);
+
+	if (source->actions_set) {
+		wl_resource_post_error(source->resource,
+				       WL_DATA_SOURCE_ERROR_INVALID_ACTION_MASK,
+				       "cannot set actions more than once");
+		return;
+	}
+
+	if (dnd_actions & ~ALL_ACTIONS) {
+		wl_resource_post_error(source->resource,
+				       WL_DATA_SOURCE_ERROR_INVALID_ACTION_MASK,
+				       "invalid action mask %x", dnd_actions);
+		return;
+	}
+
+	if (source->seat) {
+		wl_resource_post_error(source->resource,
+				       WL_DATA_SOURCE_ERROR_INVALID_ACTION_MASK,
+				       "invalid action change after "
+				       "wl_data_device.start_drag");
+		return;
+	}
+
+	source->dnd_actions = dnd_actions;
+	source->actions_set = true;
+}
+
 static struct wl_data_source_interface data_source_interface = {
 	data_source_offer,
-	data_source_destroy
+	data_source_destroy,
+	data_source_set_actions
 };
 
 static void
@@ -187,7 +404,7 @@ drag_surface_configure(struct weston_drag *drag,
 		       struct weston_surface *es,
 		       int32_t sx, int32_t sy)
 {
-	struct wl_list *list;
+	struct weston_layer_entry *list;
 	float fx, fy;
 
 	assert((pointer != NULL && touch == NULL) ||
@@ -200,10 +417,10 @@ drag_surface_configure(struct weston_drag *drag,
 		else
 			list = &es->compositor->cursor_layer.view_list;
 
-		wl_list_remove(&drag->icon->layer_link);
-		wl_list_insert(list, &drag->icon->layer_link);
+		weston_layer_entry_remove(&drag->icon->layer_link);
+		weston_layer_entry_insert(list, &drag->icon->layer_link);
 		weston_view_update_transform(drag->icon);
-		empty_region(&es->pending.input);
+		pixman_region32_clear(&es->pending.input);
 	}
 
 	drag->dx += sx;
@@ -221,6 +438,13 @@ drag_surface_configure(struct weston_drag *drag,
 	weston_view_set_position(drag->icon, fx, fy);
 }
 
+static int
+pointer_drag_surface_get_label(struct weston_surface *surface,
+			       char *buf, size_t len)
+{
+	return snprintf(buf, len, "pointer drag icon");
+}
+
 static void
 pointer_drag_surface_configure(struct weston_surface *es,
 			       int32_t sx, int32_t sy)
@@ -233,6 +457,13 @@ pointer_drag_surface_configure(struct weston_surface *es,
 	drag_surface_configure(&drag->base, pointer, NULL, es, sx, sy);
 }
 
+static int
+touch_drag_surface_get_label(struct weston_surface *surface,
+			     char *buf, size_t len)
+{
+	return snprintf(buf, len, "touch drag icon");
+}
+
 static void
 touch_drag_surface_configure(struct weston_surface *es, int32_t sx, int32_t sy)
 {
@@ -259,8 +490,9 @@ weston_drag_set_focus(struct weston_drag *drag,
 			struct weston_view *view,
 			wl_fixed_t sx, wl_fixed_t sy)
 {
-	struct wl_resource *resource, *offer = NULL;
+	struct wl_resource *resource, *offer_resource = NULL;
 	struct wl_display *display = seat->compositor->wl_display;
+	struct weston_data_offer *offer;
 	uint32_t serial;
 
 	if (drag->focus && view && drag->focus->surface == view->surface) {
@@ -282,6 +514,15 @@ weston_drag_set_focus(struct weston_drag *drag,
 	    wl_resource_get_client(view->surface->resource) != drag->client)
 		return;
 
+	if (drag->data_source &&
+	    drag->data_source->offer) {
+		/* Unlink the offer from the source */
+		offer = drag->data_source->offer;
+		offer->source = NULL;
+		drag->data_source->offer = NULL;
+		wl_list_remove(&offer->source_destroy_listener.link);
+	}
+
 	resource = wl_resource_find_for_client(&seat->drag_resource_list,
 					       wl_resource_get_client(view->surface->resource));
 	if (!resource)
@@ -290,14 +531,23 @@ weston_drag_set_focus(struct weston_drag *drag,
 	serial = wl_display_next_serial(display);
 
 	if (drag->data_source) {
-		offer = weston_data_source_send_offer(drag->data_source,
-						      resource);
+		drag->data_source->accepted = false;
+		offer = weston_data_source_send_offer(drag->data_source, resource);
 		if (offer == NULL)
 			return;
+
+		data_offer_update_action(offer);
+
+		offer_resource = offer->resource;
+		if (wl_resource_get_version (offer_resource) >=
+		    WL_DATA_OFFER_SOURCE_ACTIONS_SINCE_VERSION) {
+			wl_data_offer_send_source_actions (offer_resource,
+			                                   drag->data_source->dnd_actions);
+		}
 	}
 
 	wl_data_device_send_enter(resource, serial, view->surface->resource,
-				  sx, sy, offer);
+				  sx, sy, offer_resource);
 
 	drag->focus = view;
 	drag->focus_listener.notify = destroy_drag_focus;
@@ -323,7 +573,7 @@ drag_grab_focus(struct weston_pointer_grab *grab)
 
 static void
 drag_grab_motion(struct weston_pointer_grab *grab, uint32_t time,
-		 wl_fixed_t x, wl_fixed_t y)
+		 struct weston_pointer_motion_event *event)
 {
 	struct weston_pointer_drag *drag =
 		container_of(grab, struct weston_pointer_drag, grab);
@@ -331,7 +581,7 @@ drag_grab_motion(struct weston_pointer_grab *grab, uint32_t time,
 	float fx, fy;
 	wl_fixed_t sx, sy;
 
-	weston_pointer_move(pointer, x, y);
+	weston_pointer_move(pointer, event);
 
 	if (drag->base.icon) {
 		fx = wl_fixed_to_double(pointer->x) + drag->base.dx;
@@ -358,7 +608,8 @@ data_device_end_drag_grab(struct weston_drag *drag,
 			weston_view_unmap(drag->icon);
 
 		drag->icon->surface->configure = NULL;
-		empty_region(&drag->icon->surface->pending.input);
+		weston_surface_set_label_func(drag->icon->surface, NULL);
+		pixman_region32_clear(&drag->icon->surface->pending.input);
 		wl_list_remove(&drag->icon_destroy_listener.link);
 		weston_view_destroy(drag->icon);
 	}
@@ -370,9 +621,11 @@ static void
 data_device_end_pointer_drag_grab(struct weston_pointer_drag *drag)
 {
 	struct weston_pointer *pointer = drag->grab.pointer;
+	struct weston_keyboard *keyboard = drag->base.keyboard_grab.keyboard;
 
 	data_device_end_drag_grab(&drag->base, pointer->seat);
 	weston_pointer_end_grab(pointer);
+	weston_keyboard_end_grab(keyboard);
 	free(drag);
 }
 
@@ -384,11 +637,30 @@ drag_grab_button(struct weston_pointer_grab *grab,
 		container_of(grab, struct weston_pointer_drag, grab);
 	struct weston_pointer *pointer = drag->grab.pointer;
 	enum wl_pointer_button_state state = state_w;
+	struct weston_data_source *data_source = drag->base.data_source;
 
-	if (drag->base.focus_resource &&
+	if (data_source &&
 	    pointer->grab_button == button &&
-	    state == WL_POINTER_BUTTON_STATE_RELEASED)
-		wl_data_device_send_drop(drag->base.focus_resource);
+	    state == WL_POINTER_BUTTON_STATE_RELEASED) {
+		if (drag->base.focus_resource &&
+		    data_source->accepted &&
+		    data_source->current_dnd_action) {
+			wl_data_device_send_drop(drag->base.focus_resource);
+
+			if (wl_resource_get_version(data_source->resource) >=
+			    WL_DATA_SOURCE_DND_DROP_PERFORMED_SINCE_VERSION)
+				wl_data_source_send_dnd_drop_performed(data_source->resource);
+
+			data_source->offer->in_ask =
+				data_source->current_dnd_action ==
+				WL_DATA_DEVICE_MANAGER_DND_ACTION_ASK;
+
+			data_source->seat = NULL;
+		} else if (wl_resource_get_version(data_source->resource) >=
+			   WL_DATA_SOURCE_DND_FINISHED_SINCE_VERSION) {
+			wl_data_source_send_cancelled(data_source->resource);
+		}
+	}
 
 	if (pointer->button_count == 0 &&
 	    state == WL_POINTER_BUTTON_STATE_RELEASED) {
@@ -398,6 +670,22 @@ drag_grab_button(struct weston_pointer_grab *grab,
 	}
 }
 
+static void
+drag_grab_axis(struct weston_pointer_grab *grab,
+	       uint32_t time, struct weston_pointer_axis_event *event)
+{
+}
+
+static void
+drag_grab_axis_source(struct weston_pointer_grab *grab, uint32_t source)
+{
+}
+
+static void
+drag_grab_frame(struct weston_pointer_grab *grab)
+{
+}
+
 static void
 drag_grab_cancel(struct weston_pointer_grab *grab)
 {
@@ -414,6 +702,9 @@ static const struct weston_pointer_grab_interface pointer_drag_grab_interface =
 	drag_grab_focus,
 	drag_grab_motion,
 	drag_grab_button,
+	drag_grab_axis,
+	drag_grab_axis_source,
+	drag_grab_frame,
 	drag_grab_cancel,
 };
 
@@ -427,9 +718,11 @@ static void
 data_device_end_touch_drag_grab(struct weston_touch_drag *drag)
 {
 	struct weston_touch *touch = drag->grab.touch;
+	struct weston_keyboard *keyboard = drag->base.keyboard_grab.keyboard;
 
 	data_device_end_drag_grab(&drag->base, touch->seat);
 	weston_touch_end_grab(touch);
+	weston_keyboard_end_grab(keyboard);
 	free(drag);
 }
 
@@ -468,7 +761,7 @@ drag_grab_touch_focus(struct weston_touch_drag *drag)
 
 static void
 drag_grab_touch_motion(struct weston_touch_grab *grab, uint32_t time,
-		int touch_id, wl_fixed_t sx, wl_fixed_t sy)
+		int touch_id, wl_fixed_t x, wl_fixed_t y)
 {
 	struct weston_touch_drag *touch_drag =
 		container_of(grab, struct weston_touch_drag, grab);
@@ -496,6 +789,11 @@ drag_grab_touch_motion(struct weston_touch_grab *grab, uint32_t time,
 	}
 }
 
+static void
+drag_grab_touch_frame(struct weston_touch_grab *grab)
+{
+}
+
 static void
 drag_grab_touch_cancel(struct weston_touch_grab *grab)
 {
@@ -511,9 +809,65 @@ static const struct weston_touch_grab_interface touch_drag_grab_interface = {
 	drag_grab_touch_down,
 	drag_grab_touch_up,
 	drag_grab_touch_motion,
+	drag_grab_touch_frame,
 	drag_grab_touch_cancel
 };
 
+static void
+drag_grab_keyboard_key(struct weston_keyboard_grab *grab,
+		       uint32_t time, uint32_t key, uint32_t state)
+{
+}
+
+static void
+drag_grab_keyboard_modifiers(struct weston_keyboard_grab *grab,
+			     uint32_t serial, uint32_t mods_depressed,
+			     uint32_t mods_latched,
+			     uint32_t mods_locked, uint32_t group)
+{
+	struct weston_keyboard *keyboard = grab->keyboard;
+	struct weston_drag *drag =
+		container_of(grab, struct weston_drag, keyboard_grab);
+	uint32_t compositor_action;
+
+	if (mods_depressed & (1 << keyboard->xkb_info->shift_mod))
+		compositor_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_MOVE;
+	else if (mods_depressed & (1 << keyboard->xkb_info->ctrl_mod))
+		compositor_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_COPY;
+	else
+		compositor_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
+
+	drag->data_source->compositor_action = compositor_action;
+
+	if (drag->data_source->offer)
+		data_offer_update_action(drag->data_source->offer);
+}
+
+static void
+drag_grab_keyboard_cancel(struct weston_keyboard_grab *grab)
+{
+	struct weston_drag *drag =
+		container_of(grab, struct weston_drag, keyboard_grab);
+	struct weston_pointer *pointer = grab->keyboard->seat->pointer_state;
+	struct weston_touch *touch = grab->keyboard->seat->touch_state;
+
+	if (pointer && pointer->grab->interface == &pointer_drag_grab_interface) {
+		struct weston_touch_drag *touch_drag =
+			(struct weston_touch_drag *) drag;
+		drag_grab_touch_cancel(&touch_drag->grab);
+	} else if (touch && touch->grab->interface == &touch_drag_grab_interface) {
+		struct weston_pointer_drag *pointer_drag =
+			(struct weston_pointer_drag *) drag;
+		drag_grab_cancel(&pointer_drag->grab);
+	}
+}
+
+static const struct weston_keyboard_grab_interface keyboard_drag_grab_interface = {
+	drag_grab_keyboard_key,
+	drag_grab_keyboard_modifiers,
+	drag_grab_keyboard_cancel
+};
+
 static void
 destroy_pointer_data_device_source(struct wl_listener *listener, void *data)
 {
@@ -539,12 +893,15 @@ weston_pointer_start_drag(struct weston_pointer *pointer,
 		       struct wl_client *client)
 {
 	struct weston_pointer_drag *drag;
+	struct weston_keyboard *keyboard =
+		weston_seat_get_keyboard(pointer->seat);
 
 	drag = zalloc(sizeof *drag);
 	if (drag == NULL)
 		return -1;
 
 	drag->grab.interface = &pointer_drag_grab_interface;
+	drag->base.keyboard_grab.interface = &keyboard_drag_grab_interface;
 	drag->base.client = client;
 	drag->base.data_source = source;
 
@@ -561,6 +918,8 @@ weston_pointer_start_drag(struct weston_pointer *pointer,
 
 		icon->configure = pointer_drag_surface_configure;
 		icon->configure_private = drag;
+		weston_surface_set_label_func(icon,
+					pointer_drag_surface_get_label);
 	} else {
 		drag->base.icon = NULL;
 	}
@@ -571,9 +930,11 @@ weston_pointer_start_drag(struct weston_pointer *pointer,
 			      &drag->base.data_source_listener);
 	}
 
-	weston_pointer_set_focus(pointer, NULL,
-				 wl_fixed_from_int(0), wl_fixed_from_int(0));
+	weston_pointer_clear_focus(pointer);
+	weston_keyboard_set_focus(keyboard, NULL);
+
 	weston_pointer_start_grab(pointer, &drag->grab);
+	weston_keyboard_start_grab(keyboard, &drag->base.keyboard_grab);
 
 	return 0;
 }
@@ -594,6 +955,8 @@ weston_touch_start_drag(struct weston_touch *touch,
 		       struct wl_client *client)
 {
 	struct weston_touch_drag *drag;
+	struct weston_keyboard *keyboard =
+		weston_seat_get_keyboard(touch->seat);
 
 	drag = zalloc(sizeof *drag);
 	if (drag == NULL)
@@ -616,6 +979,8 @@ weston_touch_start_drag(struct weston_touch *touch,
 
 		icon->configure = touch_drag_surface_configure;
 		icon->configure_private = drag;
+		weston_surface_set_label_func(icon,
+					touch_drag_surface_get_label);
 	} else {
 		drag->base.icon = NULL;
 	}
@@ -626,7 +991,10 @@ weston_touch_start_drag(struct weston_touch *touch,
 			      &drag->base.data_source_listener);
 	}
 
+	weston_keyboard_set_focus(keyboard, NULL);
+
 	weston_touch_start_grab(touch, &drag->grab);
+	weston_keyboard_start_grab(keyboard, &drag->base.keyboard_grab);
 
 	drag_grab_touch_focus(drag);
 
@@ -640,17 +1008,27 @@ data_device_start_drag(struct wl_client *client, struct wl_resource *resource,
 		       struct wl_resource *icon_resource, uint32_t serial)
 {
 	struct weston_seat *seat = wl_resource_get_user_data(resource);
+	struct weston_pointer *pointer = weston_seat_get_pointer(seat);
+	struct weston_touch *touch = weston_seat_get_touch(seat);
+	struct weston_surface *origin = wl_resource_get_user_data(origin_resource);
 	struct weston_data_source *source = NULL;
 	struct weston_surface *icon = NULL;
+	int is_pointer_grab, is_touch_grab;
 	int32_t ret = 0;
 
-	if ((seat->pointer->button_count == 0 ||
-	    seat->pointer->grab_serial != serial ||
-	    !seat->pointer->focus ||
-	    seat->pointer->focus->surface != wl_resource_get_user_data(origin_resource)) &&
-		(seat->touch->grab_serial != serial ||
-		!seat->touch->focus ||
-		seat->touch->focus->surface != wl_resource_get_user_data(origin_resource)))
+	is_pointer_grab = pointer &&
+			  pointer->button_count == 1 &&
+			  pointer->grab_serial == serial &&
+			  pointer->focus &&
+			  pointer->focus->surface == origin;
+
+	is_touch_grab = touch &&
+			touch->num_tp == 1 &&
+			touch->grab_serial == serial &&
+			touch->focus &&
+			touch->focus->surface == origin;
+
+	if (!is_pointer_grab && !is_touch_grab)
 		return;
 
 	/* FIXME: Check that the data source type array isn't empty. */
@@ -659,25 +1037,23 @@ data_device_start_drag(struct wl_client *client, struct wl_resource *resource,
 		source = wl_resource_get_user_data(source_resource);
 	if (icon_resource)
 		icon = wl_resource_get_user_data(icon_resource);
-	if (icon && icon->configure) {
-		wl_resource_post_error(icon_resource,
-				       WL_DISPLAY_ERROR_INVALID_OBJECT,
-				       "surface->configure already set");
-		return;
+
+	if (icon) {
+		if (weston_surface_set_role(icon, "wl_data_device-icon",
+					    resource,
+					    WL_DATA_DEVICE_ERROR_ROLE) < 0)
+			return;
 	}
 
-	if (seat->pointer->button_count == 1 &&
-		seat->pointer->grab_serial == serial &&
-		seat->pointer->focus &&
-		seat->pointer->focus->surface == wl_resource_get_user_data(origin_resource))
-		ret = weston_pointer_start_drag(seat->pointer, source, icon, client);
-	else if (seat->touch->grab_serial != serial ||
-		seat->touch->focus ||
-		seat->touch->focus->surface != wl_resource_get_user_data(origin_resource))
-		ret = weston_touch_start_drag(seat->touch, source, icon, client);
+	if (is_pointer_grab)
+		ret = weston_pointer_start_drag(pointer, source, icon, client);
+	else if (is_touch_grab)
+		ret = weston_touch_start_drag(touch, source, icon, client);
 
 	if (ret < 0)
 		wl_resource_post_no_memory(resource);
+	else
+		source->seat = seat;
 }
 
 static void
@@ -685,13 +1061,14 @@ destroy_selection_data_source(struct wl_listener *listener, void *data)
 {
 	struct weston_seat *seat = container_of(listener, struct weston_seat,
 						selection_data_source_listener);
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
 	struct wl_resource *data_device;
 	struct weston_surface *focus = NULL;
 
 	seat->selection_data_source = NULL;
 
-	if (seat->keyboard)
-		focus = seat->keyboard->focus;
+	if (keyboard)
+		focus = keyboard->focus;
 	if (focus && focus->resource) {
 		data_device = wl_resource_find_for_client(&seat->drag_resource_list,
 							  wl_resource_get_client(focus->resource));
@@ -702,12 +1079,46 @@ destroy_selection_data_source(struct wl_listener *listener, void *data)
 	wl_signal_emit(&seat->selection_signal, seat);
 }
 
+/** \brief Send the selection to the specified client
+ *
+ * This function creates a new wl_data_offer if there is a wl_data_source
+ * currently set as the selection and sends it to the specified client,
+ * followed by the wl_data_device.selection() event.
+ * If there is no current selection the wl_data_device.selection() event
+ * will carry a NULL wl_data_offer.
+ *
+ * If the client does not have a wl_data_device for the specified seat
+ * nothing will be done.
+ *
+ * \param seat The seat owning the wl_data_device used to send the events.
+ * \param client The client to which to send the selection.
+ */
+WL_EXPORT void
+weston_seat_send_selection(struct weston_seat *seat, struct wl_client *client)
+{
+	struct weston_data_offer *offer;
+	struct wl_resource *data_device;
+
+	wl_resource_for_each(data_device, &seat->drag_resource_list) {
+		if (wl_resource_get_client(data_device) != client)
+		    continue;
+
+		if (seat->selection_data_source) {
+			offer = weston_data_source_send_offer(seat->selection_data_source,
+							      data_device);
+			wl_data_device_send_selection(data_device, offer->resource);
+		} else {
+			wl_data_device_send_selection(data_device, NULL);
+		}
+	}
+}
+
 WL_EXPORT void
 weston_seat_set_selection(struct weston_seat *seat,
 			  struct weston_data_source *source, uint32_t serial)
 {
-	struct wl_resource *data_device, *offer;
 	struct weston_surface *focus = NULL;
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
 
 	if (seat->selection_data_source &&
 	    seat->selection_serial - serial < UINT32_MAX / 2)
@@ -722,18 +1133,10 @@ weston_seat_set_selection(struct weston_seat *seat,
 	seat->selection_data_source = source;
 	seat->selection_serial = serial;
 
-	if (seat->keyboard)
-		focus = seat->keyboard->focus;
+	if (keyboard)
+		focus = keyboard->focus;
 	if (focus && focus->resource) {
-		data_device = wl_resource_find_for_client(&seat->drag_resource_list,
-							  wl_resource_get_client(focus->resource));
-		if (data_device && source) {
-			offer = weston_data_source_send_offer(seat->selection_data_source,
-							      data_device);
-			wl_data_device_send_selection(data_device, offer);
-		} else if (data_device) {
-			wl_data_device_send_selection(data_device, NULL);
-		}
+		weston_seat_send_selection(seat, wl_resource_get_client(focus->resource));
 	}
 
 	wl_signal_emit(&seat->selection_signal, seat);
@@ -751,18 +1154,34 @@ data_device_set_selection(struct wl_client *client,
 			  struct wl_resource *resource,
 			  struct wl_resource *source_resource, uint32_t serial)
 {
+	struct weston_data_source *source;
+
 	if (!source_resource)
 		return;
 
+	source = wl_resource_get_user_data(source_resource);
+
+	if (source->actions_set) {
+		wl_resource_post_error(source_resource,
+				       WL_DATA_SOURCE_ERROR_INVALID_SOURCE,
+				       "cannot set drag-and-drop source as selection");
+		return;
+	}
+
 	/* FIXME: Store serial and check against incoming serial here. */
 	weston_seat_set_selection(wl_resource_get_user_data(resource),
-				  wl_resource_get_user_data(source_resource),
-				  serial);
+				  source, serial);
+}
+static void
+data_device_release(struct wl_client *client, struct wl_resource *resource)
+{
+	wl_resource_destroy(resource);
 }
 
 static const struct wl_data_device_interface data_device_interface = {
 	data_device_start_drag,
 	data_device_set_selection,
+	data_device_release
 };
 
 static void
@@ -815,15 +1234,29 @@ create_data_source(struct wl_client *client,
 		return;
 	}
 
+	source->resource =
+		wl_resource_create(client, &wl_data_source_interface,
+				   wl_resource_get_version(resource), id);
+	if (source->resource == NULL) {
+		free(source);
+		wl_resource_post_no_memory(resource);
+		return;
+	}
+
 	wl_signal_init(&source->destroy_signal);
 	source->accept = client_source_accept;
 	source->send = client_source_send;
 	source->cancel = client_source_cancel;
+	source->offer = NULL;
+	source->accepted = false;
+	source->seat = NULL;
+	source->actions_set = false;
+	source->dnd_actions = 0;
+	source->current_dnd_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
+	source->compositor_action = WL_DATA_DEVICE_MANAGER_DND_ACTION_NONE;
 
 	wl_array_init(&source->mime_types);
 
-	source->resource =
-		wl_resource_create(client, &wl_data_source_interface, 1, id);
 	wl_resource_set_implementation(source->resource, &data_source_interface,
 				       source, destroy_data_source);
 }
@@ -842,7 +1275,9 @@ get_data_device(struct wl_client *client,
 	struct wl_resource *resource;
 
 	resource = wl_resource_create(client,
-				      &wl_data_device_interface, 1, id);
+				      &wl_data_device_interface,
+				      wl_resource_get_version(manager_resource),
+				      id);
 	if (resource == NULL) {
 		wl_resource_post_no_memory(manager_resource);
 		return;
@@ -865,9 +1300,9 @@ bind_manager(struct wl_client *client,
 {
 	struct wl_resource *resource;
 
-	resource =
-		wl_resource_create(client,
-				   &wl_data_device_manager_interface, 1, id);
+	resource = wl_resource_create(client,
+				      &wl_data_device_manager_interface,
+				      version, id);
 	if (resource == NULL) {
 		wl_client_post_no_memory(client);
 		return;
@@ -880,34 +1315,24 @@ bind_manager(struct wl_client *client,
 WL_EXPORT void
 wl_data_device_set_keyboard_focus(struct weston_seat *seat)
 {
-	struct wl_resource *data_device, *offer;
-	struct weston_data_source *source;
 	struct weston_surface *focus;
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
 
-	if (!seat->keyboard)
+	if (!keyboard)
 		return;
 
-	focus = seat->keyboard->focus;
+	focus = keyboard->focus;
 	if (!focus || !focus->resource)
 		return;
 
-	data_device = wl_resource_find_for_client(&seat->drag_resource_list,
-						  wl_resource_get_client(focus->resource));
-	if (!data_device)
-		return;
-
-	source = seat->selection_data_source;
-	if (source) {
-		offer = weston_data_source_send_offer(source, data_device);
-		wl_data_device_send_selection(data_device, offer);
-	}
+	weston_seat_send_selection(seat, wl_resource_get_client(focus->resource));
 }
 
 WL_EXPORT int
 wl_data_device_manager_init(struct wl_display *display)
 {
 	if (wl_global_create(display,
-			     &wl_data_device_manager_interface, 1,
+			     &wl_data_device_manager_interface, 3,
 			     NULL, bind_manager) == NULL)
 		return -1;
 
