@@ -1,23 +1,26 @@
 /*
  * Copyright © 2012-2013 Raspberry Pi Foundation
  *
- * Permission to use, copy, modify, distribute, and sell this software and
- * its documentation for any purpose is hereby granted without fee, provided
- * that the above copyright notice appear in all copies and that both that
- * copyright notice and this permission notice appear in supporting
- * documentation, and that the name of the copyright holders not be used in
- * advertising or publicity pertaining to distribution of the software
- * without specific, written prior permission.  The copyright holders make
- * no representations about the suitability of this software for any
- * purpose.  It is provided "as is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
- * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
- * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
- * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
- * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
- * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
- * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include "config.h"
@@ -34,6 +37,7 @@
 
 #include "compositor.h"
 #include "rpi-renderer.h"
+#include "shared/helpers.h"
 
 #ifdef ENABLE_EGL
 #include <EGL/egl.h>
@@ -149,7 +153,6 @@ struct rpir_output {
 	DISPMANX_DISPLAY_HANDLE_T display;
 
 	DISPMANX_UPDATE_HANDLE_T update;
-	struct weston_matrix matrix;
 
 	/* all Elements currently on screen */
 	struct wl_list view_list; /* struct rpir_surface::link */
@@ -337,8 +340,7 @@ apply_opaque_region(struct wl_shm_buffer *buffer,
 	for (y = 0; y < height; y++) {
 		for (x = 0; x < width; x++) {
 			int i = y * stride / 4 + x;
-			pixman_box32_t box;
-			if (pixman_region32_contains_point (opaque_region, x, y, &box)) {
+			if (pixman_region32_contains_point (opaque_region, x, y, NULL)) {
 				dst[i] = src[i] | 0xff000000;
 			} else {
 				dst[i] = src[i];
@@ -492,12 +494,11 @@ rpir_surface_create(struct rpi_renderer *renderer)
 {
 	struct rpir_surface *surface;
 
-	surface = calloc(1, sizeof *surface);
-	if (!surface)
+	surface = zalloc(sizeof *surface);
+	if (surface == NULL)
 		return NULL;
 
 	wl_list_init(&surface->views);
-	surface->visible_views = 0;
 	surface->single_buffer = renderer->single_buffer;
 	surface->enable_opaque_regions = renderer->enable_opaque_regions;
 	rpi_resource_init(&surface->resources[0]);
@@ -577,8 +578,8 @@ rpir_view_create(struct rpir_surface *surface)
 {
 	struct rpir_view *view;
 
-	view = calloc(1, sizeof *view);
-	if (!view)
+	view = zalloc(sizeof *view);
+	if (view == NULL)
 		return NULL;
 
 	view->surface = surface;
@@ -676,7 +677,6 @@ rpir_view_compute_rects(struct rpir_view *view,
 			VC_IMAGE_TRANSFORM_T *flipmask)
 {
 	struct weston_output *output_base = view->view->surface->output;
-	struct rpir_output *output = to_rpir_output(output_base);
 	struct weston_matrix matrix = view->view->transform.matrix;
 	VC_IMAGE_TRANSFORM_T flipt = 0;
 	int src_x, src_y;
@@ -703,6 +703,9 @@ rpir_view_compute_rects(struct rpir_view *view,
 		struct weston_buffer *buffer =
 			view->surface->egl_front->buffer_ref.buffer;
 
+		if (!buffer)
+			return -1;
+
 		src_width = buffer->width << 16;
 		src_height = buffer->height << 16;
 	} else {
@@ -710,14 +713,14 @@ rpir_view_compute_rects(struct rpir_view *view,
 		src_height = view->surface->front->height << 16;
 	}
 
-	weston_matrix_multiply(&matrix, &output->matrix);
+	weston_matrix_multiply(&matrix, &output_base->matrix);
 
 #ifdef SURFACE_TRANSFORM
 	if (matrix.type >= WESTON_MATRIX_TRANSFORM_OTHER) {
 #else
 	if (matrix.type >= WESTON_MATRIX_TRANSFORM_ROTATE) {
 #endif
-		warn_bad_matrix(&matrix, &output->matrix,
+		warn_bad_matrix(&matrix, &output_base->matrix,
 				&view->view->transform.matrix);
 	} else {
 		if (matrix.type & WESTON_MATRIX_TRANSFORM_ROTATE) {
@@ -728,7 +731,7 @@ rpir_view_compute_rects(struct rpir_view *view,
 				   fabsf(matrix.d[4]) < 1e-4) {
 				/* no transpose */
 			} else {
-				warn_bad_matrix(&matrix, &output->matrix,
+				warn_bad_matrix(&matrix, &output_base->matrix,
 					&view->view->transform.matrix);
 			}
 		}
@@ -858,8 +861,8 @@ rpir_view_compute_rects(struct rpir_view *view,
 	src_height = int_max(src_height, 0);
 
 	DBG("rpir_view %p %dx%d: p1 %f, %f; p2 %f, %f\n", view,
-	    view->view->geometry.width,
-	    view->view->geometry.height,
+	    view->view->surface->width,
+	    view->view->surface->height,
 	    p1.f[0], p1.f[1], p2.f[0], p2.f[1]);
 	DBG("src rect %d;%d, %d;%d, %d;%dx%d;%d\n",
 	    src_x >> 16, src_x & 0xffff,
@@ -1322,66 +1325,6 @@ rpir_output_dmx_remove_all(struct rpir_output *output,
 	}
 }
 
-static void
-output_compute_matrix(struct weston_output *base)
-{
-	struct rpir_output *output = to_rpir_output(base);
-	struct weston_matrix *matrix = &output->matrix;
-	const float half_w = 0.5f * base->width;
-	const float half_h = 0.5f * base->height;
-	float mag;
-	float dx, dy;
-
-	weston_matrix_init(matrix);
-	weston_matrix_translate(matrix, -base->x, -base->y, 0.0f);
-
-#ifdef SURFACE_TRANSFORM
-	weston_matrix_translate(matrix, -half_w, -half_h, 0.0f);
-	switch (base->transform) {
-	case WL_OUTPUT_TRANSFORM_FLIPPED:
-		weston_matrix_scale(matrix, -1.0f, 1.0f, 1.0f);
-	case WL_OUTPUT_TRANSFORM_NORMAL:
-		/* weston_matrix_rotate_xy(matrix, 1.0f, 0.0f); no-op */
-		weston_matrix_translate(matrix, half_w, half_h, 0.0f);
-		break;
-
-	case WL_OUTPUT_TRANSFORM_FLIPPED_90:
-		weston_matrix_scale(matrix, -1.0f, 1.0f, 1.0f);
-	case WL_OUTPUT_TRANSFORM_90:
-		weston_matrix_rotate_xy(matrix, 0.0f, 1.0f);
-		weston_matrix_translate(matrix, half_h, half_w, 0.0f);
-		break;
-
-	case WL_OUTPUT_TRANSFORM_FLIPPED_180:
-		weston_matrix_scale(matrix, -1.0f, 1.0f, 1.0f);
-	case WL_OUTPUT_TRANSFORM_180:
-		weston_matrix_rotate_xy(matrix, -1.0f, 0.0f);
-		weston_matrix_translate(matrix, half_w, half_h, 0.0f);
-		break;
-
-	case WL_OUTPUT_TRANSFORM_FLIPPED_270:
-		weston_matrix_scale(matrix, -1.0f, 1.0f, 1.0f);
-	case WL_OUTPUT_TRANSFORM_270:
-		weston_matrix_rotate_xy(matrix, 0.0f, -1.0f);
-		weston_matrix_translate(matrix, half_h, half_w, 0.0f);
-		break;
-
-	default:
-		break;
-	}
-#endif
-
-	if (base->zoom.active) {
-		/* The base->zoom stuff is in GL coordinate system */
-		mag = 1.0f / (1.0f - base->zoom.spring_z.current);
-		dx = -(base->zoom.trans_x + 1.0f) * half_w;
-		dy = -(base->zoom.trans_y + 1.0f) * half_h;
-		weston_matrix_translate(matrix, dx, dy, 0.0f);
-		weston_matrix_scale(matrix, mag, mag, 1.0f);
-		weston_matrix_translate(matrix, half_w, half_h, 0.0f);
-	}
-}
-
 /* Note: this won't work right for multiple outputs. A DispmanX Element
  * is tied to one DispmanX Display, i.e. output.
  */
@@ -1398,21 +1341,19 @@ rpi_renderer_repaint_output(struct weston_output *base,
 
 	assert(output->update != DISPMANX_NO_HANDLE);
 
-	output_compute_matrix(base);
-
 	rpi_resource_release(&output->capture_buffer);
 	free(output->capture_data);
 	output->capture_data = NULL;
 
 	/* Swap resources on surfaces as needed */
 	wl_list_for_each_reverse(wv, &compositor->view_list, link)
-		wv->surface->touched = 0;
+		wv->surface->touched = false;
 
 	wl_list_for_each_reverse(wv, &compositor->view_list, link) {
 		view = to_rpir_view(wv);
 
 		if (!wv->surface->touched) {
-			wv->surface->touched = 1;
+			wv->surface->touched = true;
 
 			if (view->surface->buffer_type == BUFFER_TYPE_EGL ||
 			    view->surface->need_swap)
@@ -1469,6 +1410,7 @@ rpi_renderer_repaint_output(struct weston_output *base,
 	/* The frame_signal is emitted in rpi_renderer_finish_frame(),
 	 * so that the firmware can capture the up-to-date contents.
 	 */
+	pixman_region32_copy(&base->previous_damage, output_damage);
 }
 
 static void
@@ -1549,8 +1491,8 @@ rpi_renderer_attach(struct weston_surface *base, struct weston_buffer *buffer)
 
 		surface->buffer_type = BUFFER_TYPE_EGL;
 
-		if(surface->egl_back == NULL)
-			surface->egl_back = calloc(1, sizeof *surface->egl_back);
+		if (surface->egl_back == NULL)
+			surface->egl_back = zalloc(sizeof *surface->egl_back);
 
 		weston_buffer_reference(&surface->egl_back->buffer_ref, buffer);
 		surface->egl_back->resource_handle =
@@ -1726,7 +1668,7 @@ rpi_renderer_create(struct weston_compositor *compositor,
 
 	weston_log("Initializing the DispmanX compositing renderer\n");
 
-	renderer = calloc(1, sizeof *renderer);
+	renderer = zalloc(sizeof *renderer);
 	if (renderer == NULL)
 		return -1;
 
@@ -1798,8 +1740,8 @@ rpi_renderer_output_create(struct weston_output *base,
 
 	assert(base->renderer_state == NULL);
 
-	output = calloc(1, sizeof *output);
-	if (!output)
+	output = zalloc(sizeof *output);
+	if (output == NULL)
 		return -1;
 
 	output->display = display;
