@@ -2,36 +2,42 @@
  * Copyright © 2012 Openismus GmbH
  * Copyright © 2012 Intel Corporation
  *
- * Permission to use, copy, modify, distribute, and sell this software and
- * its documentation for any purpose is hereby granted without fee, provided
- * that the above copyright notice appear in all copies and that both that
- * copyright notice and this permission notice appear in supporting
- * documentation, and that the name of the copyright holders not be used in
- * advertising or publicity pertaining to distribution of the software
- * without specific, written prior permission.  The copyright holders make
- * no representations about the suitability of this software for any
- * purpose.  It is provided "as is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
- * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
- * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
- * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
- * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
- * CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
- * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include "config.h"
 
+#include <stdbool.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <time.h>
 
 #include "compositor.h"
-#include "text-server-protocol.h"
-#include "input-method-server-protocol.h"
+#include "text-input-unstable-v1-server-protocol.h"
+#include "input-method-unstable-v1-server-protocol.h"
+#include "shared/helpers.h"
 
+struct text_input_manager;
 struct input_method;
 struct input_method_context;
 struct text_backend;
@@ -47,13 +53,17 @@ struct text_input {
 
 	pixman_box32_t cursor_rectangle;
 
-	uint32_t input_panel_visible;
+	bool input_panel_visible;
+
+	struct text_input_manager *manager;
 };
 
 struct text_input_manager {
 	struct wl_global *text_input_manager_global;
 	struct wl_listener destroy_listener;
 
+	struct text_input *current_panel;
+
 	struct weston_compositor *ec;
 };
 
@@ -63,13 +73,13 @@ struct input_method {
 	struct wl_listener destroy_listener;
 
 	struct weston_seat *seat;
-	struct text_input *model;
+	struct text_input *input;
 
 	struct wl_list link;
 
 	struct wl_listener keyboard_focus_listener;
 
-	int focus_listener_initialized;
+	bool focus_listener_initialized;
 
 	struct input_method_context *context;
 
@@ -79,11 +89,9 @@ struct input_method {
 struct input_method_context {
 	struct wl_resource *resource;
 
-	struct text_input *model;
+	struct text_input *input;
 	struct input_method *input_method;
 
-	struct wl_list link;
-
 	struct wl_resource *keyboard;
 };
 
@@ -92,43 +100,49 @@ struct text_backend {
 
 	struct {
 		char *path;
-		struct wl_resource *binding;
-		struct weston_process process;
 		struct wl_client *client;
 
 		unsigned deathcount;
 		uint32_t deathstamp;
 	} input_method;
 
+	struct wl_listener client_listener;
 	struct wl_listener seat_created_listener;
-	struct wl_listener destroy_listener;
 };
 
-static void input_method_context_create(struct text_input *model,
-					struct input_method *input_method);
-static void input_method_context_end_keyboard_grab(struct input_method_context *context);
-
-static void input_method_init_seat(struct weston_seat *seat);
+static void
+input_method_context_create(struct text_input *input,
+			    struct input_method *input_method);
+static void
+input_method_context_end_keyboard_grab(struct input_method_context *context);
 
 static void
-deactivate_text_input(struct text_input *text_input,
-		      struct input_method *input_method)
+input_method_init_seat(struct weston_seat *seat);
+
+static void
+deactivate_input_method(struct input_method *input_method)
 {
+	struct text_input *text_input = input_method->input;
 	struct weston_compositor *ec = text_input->ec;
 
-	if (input_method->model == text_input) {
-		if (input_method->context && input_method->input_method_binding) {
-			input_method_context_end_keyboard_grab(input_method->context);
-			wl_input_method_send_deactivate(input_method->input_method_binding,
-							input_method->context->resource);
-		}
-
-		wl_list_remove(&input_method->link);
-		input_method->model = NULL;
-		input_method->context = NULL;
-		wl_signal_emit(&ec->hide_input_panel_signal, ec);
-		wl_text_input_send_leave(text_input->resource);
+	if (input_method->context && input_method->input_method_binding) {
+		input_method_context_end_keyboard_grab(input_method->context);
+		zwp_input_method_v1_send_deactivate(
+			input_method->input_method_binding,
+			input_method->context->resource);
 	}
+
+	wl_list_remove(&input_method->link);
+	input_method->input = NULL;
+	input_method->context = NULL;
+
+	if (wl_list_empty(&text_input->input_methods) &&
+	    text_input->input_panel_visible) {
+		wl_signal_emit(&ec->hide_input_panel_signal, ec);
+		text_input->input_panel_visible = false;
+		text_input->manager->current_panel = NULL;
+	}
+	zwp_text_input_v1_send_leave(text_input->resource);
 }
 
 static void
@@ -137,8 +151,9 @@ destroy_text_input(struct wl_resource *resource)
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link)
-		deactivate_text_input(text_input, input_method);
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link)
+		deactivate_input_method(input_method);
 
 	free(text_input);
 }
@@ -153,13 +168,12 @@ text_input_set_surrounding_text(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link) {
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link) {
 		if (!input_method->context)
 			continue;
-		wl_input_method_context_send_surrounding_text(input_method->context->resource,
-							      text,
-							      cursor,
-							      anchor);
+		zwp_input_method_context_v1_send_surrounding_text(
+			input_method->context->resource, text, cursor, anchor);
 	}
 }
 
@@ -172,18 +186,16 @@ text_input_activate(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct weston_seat *weston_seat = wl_resource_get_user_data(seat);
 	struct input_method *input_method = weston_seat->input_method;
-	struct text_input *old = weston_seat->input_method->model;
 	struct weston_compositor *ec = text_input->ec;
+	struct text_input *current;
 
-	if (old == text_input)
+	if (input_method->input == text_input)
 		return;
 
-	if (old) {
-		deactivate_text_input(old,
-				      weston_seat->input_method);
-	}
+	if (input_method->input)
+		deactivate_input_method(input_method);
 
-	input_method->model = text_input;
+	input_method->input = text_input;
 	wl_list_insert(&text_input->input_methods, &input_method->link);
 	input_method_init_seat(weston_seat);
 
@@ -191,12 +203,24 @@ text_input_activate(struct wl_client *client,
 
 	input_method_context_create(text_input, input_method);
 
-	if (text_input->input_panel_visible) {
-		wl_signal_emit(&ec->show_input_panel_signal, text_input->surface);
-		wl_signal_emit(&ec->update_input_panel_signal, &text_input->cursor_rectangle);
+	current = text_input->manager->current_panel;
+
+	if (current && current != text_input) {
+		current->input_panel_visible = false;
+		wl_signal_emit(&ec->hide_input_panel_signal, ec);
+		text_input->manager->current_panel = NULL;
 	}
 
-	wl_text_input_send_enter(text_input->resource, text_input->surface->resource);
+	if (text_input->input_panel_visible) {
+		wl_signal_emit(&ec->show_input_panel_signal,
+			       text_input->surface);
+		wl_signal_emit(&ec->update_input_panel_signal,
+			       &text_input->cursor_rectangle);
+		text_input->manager->current_panel = text_input;
+	}
+
+	zwp_text_input_v1_send_enter(text_input->resource,
+				     text_input->surface->resource);
 }
 
 static void
@@ -204,11 +228,10 @@ text_input_deactivate(struct wl_client *client,
 		      struct wl_resource *resource,
 		      struct wl_resource *seat)
 {
-	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct weston_seat *weston_seat = wl_resource_get_user_data(seat);
 
-	deactivate_text_input(text_input,
-			      weston_seat->input_method);
+	if (weston_seat->input_method->input)
+		deactivate_input_method(weston_seat->input_method);
 }
 
 static void
@@ -218,10 +241,12 @@ text_input_reset(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link) {
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link) {
 		if (!input_method->context)
 			continue;
-		wl_input_method_context_send_reset(input_method->context->resource);
+		zwp_input_method_context_v1_send_reset(
+			input_method->context->resource);
 	}
 }
 
@@ -241,7 +266,8 @@ text_input_set_cursor_rectangle(struct wl_client *client,
 	text_input->cursor_rectangle.x2 = x + width;
 	text_input->cursor_rectangle.y2 = y + height;
 
-	wl_signal_emit(&ec->update_input_panel_signal, &text_input->cursor_rectangle);
+	wl_signal_emit(&ec->update_input_panel_signal,
+		       &text_input->cursor_rectangle);
 }
 
 static void
@@ -253,10 +279,12 @@ text_input_set_content_type(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link) {
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link) {
 		if (!input_method->context)
 			continue;
-		wl_input_method_context_send_content_type(input_method->context->resource, hint, purpose);
+		zwp_input_method_context_v1_send_content_type(
+			input_method->context->resource, hint, purpose);
 	}
 }
 
@@ -269,10 +297,12 @@ text_input_invoke_action(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link) {
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link) {
 		if (!input_method->context)
 			continue;
-		wl_input_method_context_send_invoke_action(input_method->context->resource, button, index);
+		zwp_input_method_context_v1_send_invoke_action(
+			input_method->context->resource, button, index);
 	}
 }
 
@@ -284,10 +314,12 @@ text_input_commit_state(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link) {
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link) {
 		if (!input_method->context)
 			continue;
-		wl_input_method_context_send_commit_state(input_method->context->resource, serial);
+		zwp_input_method_context_v1_send_commit_state(
+			input_method->context->resource, serial);
 	}
 }
 
@@ -298,11 +330,13 @@ text_input_show_input_panel(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct weston_compositor *ec = text_input->ec;
 
-	text_input->input_panel_visible = 1;
+	text_input->input_panel_visible = true;
 
 	if (!wl_list_empty(&text_input->input_methods)) {
-		wl_signal_emit(&ec->show_input_panel_signal, text_input->surface);
-		wl_signal_emit(&ec->update_input_panel_signal, &text_input->cursor_rectangle);
+		wl_signal_emit(&ec->show_input_panel_signal,
+			       text_input->surface);
+		wl_signal_emit(&ec->update_input_panel_signal,
+			       &text_input->cursor_rectangle);
 	}
 }
 
@@ -313,10 +347,13 @@ text_input_hide_input_panel(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct weston_compositor *ec = text_input->ec;
 
-	text_input->input_panel_visible = 0;
+	text_input->input_panel_visible = false;
 
-	if (!wl_list_empty(&text_input->input_methods))
+	if (!wl_list_empty(&text_input->input_methods) &&
+	    text_input == text_input->manager->current_panel) {
+		text_input->manager->current_panel = NULL;
 		wl_signal_emit(&ec->hide_input_panel_signal, ec);
+	}
 }
 
 static void
@@ -327,15 +364,16 @@ text_input_set_preferred_language(struct wl_client *client,
 	struct text_input *text_input = wl_resource_get_user_data(resource);
 	struct input_method *input_method, *next;
 
-	wl_list_for_each_safe(input_method, next, &text_input->input_methods, link) {
+	wl_list_for_each_safe(input_method, next,
+			      &text_input->input_methods, link) {
 		if (!input_method->context)
 			continue;
-		wl_input_method_context_send_preferred_language(input_method->context->resource,
-								language);
+		zwp_input_method_context_v1_send_preferred_language(
+			input_method->context->resource, language);
 	}
 }
 
-static const struct wl_text_input_interface text_input_implementation = {
+static const struct zwp_text_input_v1_interface text_input_implementation = {
 	text_input_activate,
 	text_input_deactivate,
 	text_input_show_input_panel,
@@ -353,23 +391,27 @@ static void text_input_manager_create_text_input(struct wl_client *client,
 						 struct wl_resource *resource,
 						 uint32_t id)
 {
-	struct text_input_manager *text_input_manager = wl_resource_get_user_data(resource);
+	struct text_input_manager *text_input_manager =
+		wl_resource_get_user_data(resource);
 	struct text_input *text_input;
 
-	text_input = calloc(1, sizeof *text_input);
+	text_input = zalloc(sizeof *text_input);
+	if (text_input == NULL)
+		return;
 
 	text_input->resource =
-		wl_resource_create(client, &wl_text_input_interface, 1, id);
+		wl_resource_create(client, &zwp_text_input_v1_interface, 1, id);
 	wl_resource_set_implementation(text_input->resource,
 				       &text_input_implementation,
 				       text_input, destroy_text_input);
 
 	text_input->ec = text_input_manager->ec;
+	text_input->manager = text_input_manager;
 
 	wl_list_init(&text_input->input_methods);
 };
 
-static const struct wl_text_input_manager_interface text_input_manager_implementation = {
+static const struct zwp_text_input_manager_v1_interface manager_implementation = {
 	text_input_manager_create_text_input
 };
 
@@ -385,10 +427,10 @@ bind_text_input_manager(struct wl_client *client,
 	/* No checking for duplicate binding necessary.  */
 	resource =
 		wl_resource_create(client,
-				   &wl_text_input_manager_interface, 1, id);
+				   &zwp_text_input_manager_v1_interface, 1, id);
 	if (resource)
 		wl_resource_set_implementation(resource,
-					       &text_input_manager_implementation,
+					       &manager_implementation,
 					       text_input_manager, NULL);
 }
 
@@ -396,7 +438,9 @@ static void
 text_input_manager_notifier_destroy(struct wl_listener *listener, void *data)
 {
 	struct text_input_manager *text_input_manager =
-		container_of(listener, struct text_input_manager, destroy_listener);
+		container_of(listener,
+			     struct text_input_manager,
+			     destroy_listener);
 
 	wl_global_destroy(text_input_manager->text_input_manager_global);
 
@@ -408,17 +452,21 @@ text_input_manager_create(struct weston_compositor *ec)
 {
 	struct text_input_manager *text_input_manager;
 
-	text_input_manager = calloc(1, sizeof *text_input_manager);
+	text_input_manager = zalloc(sizeof *text_input_manager);
+	if (text_input_manager == NULL)
+		return;
 
 	text_input_manager->ec = ec;
 
 	text_input_manager->text_input_manager_global =
 		wl_global_create(ec->wl_display,
-				 &wl_text_input_manager_interface, 1,
+				 &zwp_text_input_manager_v1_interface, 1,
 				 text_input_manager, bind_text_input_manager);
 
-	text_input_manager->destroy_listener.notify = text_input_manager_notifier_destroy;
-	wl_signal_add(&ec->destroy_signal, &text_input_manager->destroy_listener);
+	text_input_manager->destroy_listener.notify =
+		text_input_manager_notifier_destroy;
+	wl_signal_add(&ec->destroy_signal,
+		      &text_input_manager->destroy_listener);
 }
 
 static void
@@ -437,7 +485,9 @@ input_method_context_commit_string(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_commit_string(context->model->resource, serial, text);
+	if (context->input)
+		zwp_text_input_v1_send_commit_string(context->input->resource,
+						     serial, text);
 }
 
 static void
@@ -450,7 +500,9 @@ input_method_context_preedit_string(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_preedit_string(context->model->resource, serial, text, commit);
+	if (context->input)
+		zwp_text_input_v1_send_preedit_string(context->input->resource,
+						      serial, text, commit);
 }
 
 static void
@@ -463,7 +515,9 @@ input_method_context_preedit_styling(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_preedit_styling(context->model->resource, index, length, style);
+	if (context->input)
+		zwp_text_input_v1_send_preedit_styling(context->input->resource,
+						       index, length, style);
 }
 
 static void
@@ -474,7 +528,9 @@ input_method_context_preedit_cursor(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_preedit_cursor(context->model->resource, cursor);
+	if (context->input)
+		zwp_text_input_v1_send_preedit_cursor(context->input->resource,
+						      cursor);
 }
 
 static void
@@ -486,7 +542,9 @@ input_method_context_delete_surrounding_text(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_delete_surrounding_text(context->model->resource, index, length);
+	if (context->input)
+		zwp_text_input_v1_send_delete_surrounding_text(
+			context->input->resource, index, length);
 }
 
 static void
@@ -498,7 +556,9 @@ input_method_context_cursor_position(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_cursor_position(context->model->resource, index, anchor);
+	if (context->input)
+		zwp_text_input_v1_send_cursor_position(context->input->resource,
+						       index, anchor);
 }
 
 static void
@@ -509,7 +569,9 @@ input_method_context_modifiers_map(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_modifiers_map(context->model->resource, map);
+	if (context->input)
+		zwp_text_input_v1_send_modifiers_map(context->input->resource,
+						     map);
 }
 
 static void
@@ -524,8 +586,10 @@ input_method_context_keysym(struct wl_client *client,
 	struct input_method_context *context =
 		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_keysym(context->model->resource, serial, time,
-				  sym, state, modifiers);
+	if (context->input)
+		zwp_text_input_v1_send_keysym(context->input->resource,
+					      serial, time,
+					      sym, state, modifiers);
 }
 
 static void
@@ -549,16 +613,20 @@ input_method_context_grab_key(struct weston_keyboard_grab *grab,
 	if (!keyboard->input_method_resource)
 		return;
 
-	display = wl_client_get_display(wl_resource_get_client(keyboard->input_method_resource));
+	display = wl_client_get_display(
+		wl_resource_get_client(keyboard->input_method_resource));
 	serial = wl_display_next_serial(display);
 	wl_keyboard_send_key(keyboard->input_method_resource,
 			     serial, time, key, state_w);
 }
 
 static void
-input_method_context_grab_modifier(struct weston_keyboard_grab *grab, uint32_t serial,
-				   uint32_t mods_depressed, uint32_t mods_latched,
-				   uint32_t mods_locked, uint32_t group)
+input_method_context_grab_modifier(struct weston_keyboard_grab *grab,
+				   uint32_t serial,
+				   uint32_t mods_depressed,
+				   uint32_t mods_latched,
+				   uint32_t mods_locked,
+				   uint32_t group)
 {
 	struct weston_keyboard *keyboard = grab->keyboard;
 
@@ -570,9 +638,16 @@ input_method_context_grab_modifier(struct weston_keyboard_grab *grab, uint32_t s
 				   mods_locked, group);
 }
 
+static void
+input_method_context_grab_cancel(struct weston_keyboard_grab *grab)
+{
+	weston_keyboard_end_grab(grab->keyboard);
+}
+
 static const struct weston_keyboard_grab_interface input_method_context_grab = {
 	input_method_context_grab_key,
 	input_method_context_grab_modifier,
+	input_method_context_grab_cancel,
 };
 
 static void
@@ -580,10 +655,11 @@ input_method_context_grab_keyboard(struct wl_client *client,
 				   struct wl_resource *resource,
 				   uint32_t id)
 {
-	struct input_method_context *context = wl_resource_get_user_data(resource);
+	struct input_method_context *context =
+		wl_resource_get_user_data(resource);
 	struct wl_resource *cr;
 	struct weston_seat *seat = context->input_method->seat;
-	struct weston_keyboard *keyboard = seat->keyboard;
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
 
 	cr = wl_resource_create(client, &wl_keyboard_interface, 1, id);
 	wl_resource_set_implementation(cr, NULL, context, unbind_keyboard);
@@ -591,8 +667,8 @@ input_method_context_grab_keyboard(struct wl_client *client,
 	context->keyboard = cr;
 
 	wl_keyboard_send_keymap(cr, WL_KEYBOARD_KEYMAP_FORMAT_XKB_V1,
-				seat->xkb_info->keymap_fd,
-				seat->xkb_info->keymap_size);
+				keyboard->xkb_info->keymap_fd,
+				keyboard->xkb_info->keymap_size);
 
 	if (keyboard->grab != &keyboard->default_grab) {
 		weston_keyboard_end_grab(keyboard);
@@ -609,9 +685,10 @@ input_method_context_key(struct wl_client *client,
 			 uint32_t key,
 			 uint32_t state_w)
 {
-	struct input_method_context *context = wl_resource_get_user_data(resource);
+	struct input_method_context *context =
+		wl_resource_get_user_data(resource);
 	struct weston_seat *seat = context->input_method->seat;
-	struct weston_keyboard *keyboard = seat->keyboard;
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
 	struct weston_keyboard_grab *default_grab = &keyboard->default_grab;
 
 	default_grab->interface->key(default_grab, time, key, state_w);
@@ -626,10 +703,11 @@ input_method_context_modifiers(struct wl_client *client,
 			       uint32_t mods_locked,
 			       uint32_t group)
 {
-	struct input_method_context *context = wl_resource_get_user_data(resource);
+	struct input_method_context *context =
+		wl_resource_get_user_data(resource);
 
 	struct weston_seat *seat = context->input_method->seat;
-	struct weston_keyboard *keyboard = seat->keyboard;
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
 	struct weston_keyboard_grab *default_grab = &keyboard->default_grab;
 
 	default_grab->interface->modifiers(default_grab,
@@ -644,9 +722,12 @@ input_method_context_language(struct wl_client *client,
 			      uint32_t serial,
 			      const char *language)
 {
-	struct input_method_context *context = wl_resource_get_user_data(resource);
+	struct input_method_context *context =
+		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_language(context->model->resource, serial, language);
+	if (context->input)
+		zwp_text_input_v1_send_language(context->input->resource,
+						serial, language);
 }
 
 static void
@@ -655,13 +736,16 @@ input_method_context_text_direction(struct wl_client *client,
 				    uint32_t serial,
 				    uint32_t direction)
 {
-	struct input_method_context *context = wl_resource_get_user_data(resource);
+	struct input_method_context *context =
+		wl_resource_get_user_data(resource);
 
-	wl_text_input_send_text_direction(context->model->resource, serial, direction);
+	if (context->input)
+		zwp_text_input_v1_send_text_direction(context->input->resource,
+						      serial, direction);
 }
 
 
-static const struct wl_input_method_context_interface input_method_context_implementation = {
+static const struct zwp_input_method_context_v1_interface context_implementation = {
 	input_method_context_destroy,
 	input_method_context_commit_string,
 	input_method_context_preedit_string,
@@ -681,17 +765,20 @@ static const struct wl_input_method_context_interface input_method_context_imple
 static void
 destroy_input_method_context(struct wl_resource *resource)
 {
-	struct input_method_context *context = wl_resource_get_user_data(resource);
+	struct input_method_context *context =
+		wl_resource_get_user_data(resource);
 
-	if (context->keyboard) {
+	if (context->keyboard)
 		wl_resource_destroy(context->keyboard);
-	}
+
+	if (context->input_method && context->input_method->context == context)
+		context->input_method->context = NULL;
 
 	free(context);
 }
 
 static void
-input_method_context_create(struct text_input *model,
+input_method_context_create(struct text_input *input,
 			    struct input_method *input_method)
 {
 	struct input_method_context *context;
@@ -700,38 +787,44 @@ input_method_context_create(struct text_input *model,
 	if (!input_method->input_method_binding)
 		return;
 
-	context = calloc(1, sizeof *context);
+	context = zalloc(sizeof *context);
 	if (context == NULL)
 		return;
 
 	binding = input_method->input_method_binding;
 	context->resource =
 		wl_resource_create(wl_resource_get_client(binding),
-				   &wl_input_method_context_interface, 1, 0);
+				   &zwp_input_method_context_v1_interface,
+				   1, 0);
 	wl_resource_set_implementation(context->resource,
-				       &input_method_context_implementation,
+				       &context_implementation,
 				       context, destroy_input_method_context);
 
-	context->model = model;
+	context->input = input;
 	context->input_method = input_method;
 	input_method->context = context;
 
 
-	wl_input_method_send_activate(binding, context->resource);
+	zwp_input_method_v1_send_activate(binding, context->resource);
 }
 
 static void
 input_method_context_end_keyboard_grab(struct input_method_context *context)
 {
-	struct weston_keyboard_grab *grab =
-		&context->input_method->seat->keyboard->input_method_grab;
-	struct weston_keyboard *keyboard = grab->keyboard;
+	struct weston_keyboard_grab *grab;
+	struct weston_keyboard *keyboard;
 
-	if (!grab->keyboard)
+	keyboard = weston_seat_get_keyboard(context->input_method->seat);
+	if (!keyboard)
 		return;
 
-	if (grab->keyboard->grab == grab)
-		weston_keyboard_end_grab(grab->keyboard);
+	grab = &keyboard->input_method_grab;
+	keyboard = grab->keyboard;
+	if (!keyboard)
+		return;
+
+	if (keyboard->grab == grab)
+		weston_keyboard_end_grab(keyboard);
 
 	keyboard->input_method_resource = NULL;
 }
@@ -740,12 +833,9 @@ static void
 unbind_input_method(struct wl_resource *resource)
 {
 	struct input_method *input_method = wl_resource_get_user_data(resource);
-	struct text_backend *text_backend = input_method->text_backend;
 
 	input_method->input_method_binding = NULL;
 	input_method->context = NULL;
-
-	text_backend->input_method.binding = NULL;
 }
 
 static void
@@ -759,27 +849,27 @@ bind_input_method(struct wl_client *client,
 	struct wl_resource *resource;
 
 	resource =
-		wl_resource_create(client, &wl_input_method_interface, 1, id);
+		wl_resource_create(client,
+				   &zwp_input_method_v1_interface, 1, id);
 
 	if (input_method->input_method_binding != NULL) {
-		wl_resource_post_error(resource, WL_DISPLAY_ERROR_INVALID_OBJECT,
+		wl_resource_post_error(resource,
+				       WL_DISPLAY_ERROR_INVALID_OBJECT,
 				       "interface object already bound");
-		wl_resource_destroy(resource);
 		return;
 	}
 
 	if (text_backend->input_method.client != client) {
-		wl_resource_post_error(resource, WL_DISPLAY_ERROR_INVALID_OBJECT,
-				       "permission to bind input_method denied");
-		wl_resource_destroy(resource);
+		wl_resource_post_error(resource,
+				       WL_DISPLAY_ERROR_INVALID_OBJECT,
+				       "permission to bind "
+				       "input_method denied");
 		return;
 	}
 
 	wl_resource_set_implementation(resource, NULL, input_method,
 				       unbind_input_method);
 	input_method->input_method_binding = resource;
-
-	text_backend->input_method.binding = resource;
 }
 
 static void
@@ -788,8 +878,8 @@ input_method_notifier_destroy(struct wl_listener *listener, void *data)
 	struct input_method *input_method =
 		container_of(listener, struct input_method, destroy_listener);
 
-	if (input_method->model)
-		deactivate_text_input(input_method->model, input_method);
+	if (input_method->input)
+		deactivate_input_method(input_method);
 
 	wl_global_destroy(input_method->input_method_global);
 	wl_list_remove(&input_method->destroy_listener.link);
@@ -802,43 +892,43 @@ handle_keyboard_focus(struct wl_listener *listener, void *data)
 {
 	struct weston_keyboard *keyboard = data;
 	struct input_method *input_method =
-		container_of(listener, struct input_method, keyboard_focus_listener);
+		container_of(listener, struct input_method,
+			     keyboard_focus_listener);
 	struct weston_surface *surface = keyboard->focus;
 
-	if (!input_method->model)
+	if (!input_method->input)
 		return;
 
-	if (!surface || input_method->model->surface != surface)
-		deactivate_text_input(input_method->model,
-				      input_method);
+	if (!surface || input_method->input->surface != surface)
+		deactivate_input_method(input_method);
 }
 
 static void
 input_method_init_seat(struct weston_seat *seat)
 {
+	struct weston_keyboard *keyboard = weston_seat_get_keyboard(seat);
+
 	if (seat->input_method->focus_listener_initialized)
 		return;
 
-	if (seat->keyboard) {
-		seat->input_method->keyboard_focus_listener.notify = handle_keyboard_focus;
-		wl_signal_add(&seat->keyboard->focus_signal, &seat->input_method->keyboard_focus_listener);
-		seat->keyboard->input_method_grab.interface = &input_method_context_grab;
+	if (keyboard) {
+		seat->input_method->keyboard_focus_listener.notify =
+			handle_keyboard_focus;
+		wl_signal_add(&keyboard->focus_signal,
+			      &seat->input_method->keyboard_focus_listener);
+		keyboard->input_method_grab.interface =
+			&input_method_context_grab;
 	}
 
-	seat->input_method->focus_listener_initialized = 1;
+	seat->input_method->focus_listener_initialized = true;
 }
 
 static void launch_input_method(struct text_backend *text_backend);
 
 static void
-handle_input_method_sigchld(struct weston_process *process, int status)
+respawn_input_method_process(struct text_backend *text_backend)
 {
 	uint32_t time;
-	struct text_backend *text_backend =
-		container_of(process, struct text_backend, input_method.process);
-
-	text_backend->input_method.process.pid = 0;
-	text_backend->input_method.client = NULL;
 
 	/* if input_method dies more than 5 times in 10 seconds, give up */
 	time = weston_compositor_get_time();
@@ -849,112 +939,144 @@ handle_input_method_sigchld(struct weston_process *process, int status)
 
 	text_backend->input_method.deathcount++;
 	if (text_backend->input_method.deathcount > 5) {
-		weston_log("input_method died, giving up.\n");
+		weston_log("input_method disconnected, giving up.\n");
 		return;
 	}
 
-	weston_log("input_method died, respawning...\n");
+	weston_log("input_method disconnected, respawning...\n");
 	launch_input_method(text_backend);
 }
 
+static void
+input_method_client_notifier(struct wl_listener *listener, void *data)
+{
+	struct text_backend *text_backend;
+
+	text_backend = container_of(listener, struct text_backend,
+				    client_listener);
+
+	text_backend->input_method.client = NULL;
+	respawn_input_method_process(text_backend);
+}
+
 static void
 launch_input_method(struct text_backend *text_backend)
 {
-	if (text_backend->input_method.binding)
-		return;
-
 	if (!text_backend->input_method.path)
 		return;
 
-	if (text_backend->input_method.process.pid != 0)
+	if (strcmp(text_backend->input_method.path, "") == 0)
 		return;
 
-	text_backend->input_method.client = weston_client_launch(text_backend->compositor,
-								 &text_backend->input_method.process,
-								 text_backend->input_method.path,
-								 handle_input_method_sigchld);
+	text_backend->input_method.client =
+		weston_client_start(text_backend->compositor,
+				    text_backend->input_method.path);
 
-	if (!text_backend->input_method.client)
-		weston_log("not able to start %s\n", text_backend->input_method.path);
+	if (!text_backend->input_method.client) {
+		weston_log("not able to start %s\n",
+			   text_backend->input_method.path);
+		return;
+	}
+
+	text_backend->client_listener.notify = input_method_client_notifier;
+	wl_client_add_destroy_listener(text_backend->input_method.client,
+				       &text_backend->client_listener);
 }
 
 static void
-handle_seat_created(struct wl_listener *listener,
-		    void *data)
+text_backend_seat_created(struct text_backend *text_backend,
+			  struct weston_seat *seat)
 {
-	struct weston_seat *seat = data;
-	struct text_backend *text_backend =
-		container_of(listener, struct text_backend,
-			     seat_created_listener);
 	struct input_method *input_method;
 	struct weston_compositor *ec = seat->compositor;
 
-	input_method = calloc(1, sizeof *input_method);
+	input_method = zalloc(sizeof *input_method);
+	if (input_method == NULL)
+		return;
 
 	input_method->seat = seat;
-	input_method->model = NULL;
-	input_method->focus_listener_initialized = 0;
+	input_method->input = NULL;
+	input_method->focus_listener_initialized = false;
 	input_method->context = NULL;
 	input_method->text_backend = text_backend;
 
 	input_method->input_method_global =
-		wl_global_create(ec->wl_display, &wl_input_method_interface, 1,
+		wl_global_create(ec->wl_display,
+				 &zwp_input_method_v1_interface, 1,
 				 input_method, bind_input_method);
 
 	input_method->destroy_listener.notify = input_method_notifier_destroy;
 	wl_signal_add(&seat->destroy_signal, &input_method->destroy_listener);
 
 	seat->input_method = input_method;
+}
 
-	launch_input_method(text_backend);
+static void
+handle_seat_created(struct wl_listener *listener, void *data)
+{
+	struct weston_seat *seat = data;
+	struct text_backend *text_backend =
+		container_of(listener, struct text_backend,
+			     seat_created_listener);
+
+	text_backend_seat_created(text_backend, seat);
 }
 
 static void
 text_backend_configuration(struct text_backend *text_backend)
 {
 	struct weston_config_section *section;
+	char *client;
+	int ret;
 
 	section = weston_config_get_section(text_backend->compositor->config,
 					    "input-method", NULL, NULL);
+	ret = asprintf(&client, "%s/weston-keyboard",
+		       weston_config_get_libexec_dir());
+	if (ret < 0)
+		client = NULL;
 	weston_config_section_get_string(section, "path",
 					 &text_backend->input_method.path,
-					 LIBEXECDIR "/weston-keyboard");
+					 client);
+	free(client);
 }
 
-static void
-text_backend_notifier_destroy(struct wl_listener *listener, void *data)
+WL_EXPORT void
+text_backend_destroy(struct text_backend *text_backend)
 {
-	struct text_backend *text_backend =
-		container_of(listener, struct text_backend, destroy_listener);
-
-	if (text_backend->input_method.client)
+	if (text_backend->input_method.client) {
+		/* disable respawn */
+		wl_list_remove(&text_backend->client_listener.link);
 		wl_client_destroy(text_backend->input_method.client);
+	}
 
 	free(text_backend->input_method.path);
-
 	free(text_backend);
 }
 
-
-WL_EXPORT int
+WL_EXPORT struct text_backend *
 text_backend_init(struct weston_compositor *ec)
 {
 	struct text_backend *text_backend;
+	struct weston_seat *seat;
 
-	text_backend = calloc(1, sizeof(*text_backend));
+	text_backend = zalloc(sizeof(*text_backend));
+	if (text_backend == NULL)
+		return NULL;
 
 	text_backend->compositor = ec;
 
+	text_backend_configuration(text_backend);
+
+	wl_list_for_each(seat, &ec->seat_list, link)
+		text_backend_seat_created(text_backend, seat);
 	text_backend->seat_created_listener.notify = handle_seat_created;
 	wl_signal_add(&ec->seat_created_signal,
 		      &text_backend->seat_created_listener);
 
-	text_backend->destroy_listener.notify = text_backend_notifier_destroy;
-	wl_signal_add(&ec->destroy_signal, &text_backend->destroy_listener);
-
-	text_backend_configuration(text_backend);
-
 	text_input_manager_create(ec);
 
-	return 0;
+	launch_input_method(text_backend);
+
+	return text_backend;
 }
